/* Pxys - An IRC ProxyScan, ms_parse.c
 * 
 * Copyright (C) 2001 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * Server messages parsing.
 *
 * $Id: ms_parser.c,v 1.14 2001/11/21 15:43:23 mbuna Exp $
 */

#include "config.h"


#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "m_array.h"
#include "m_log.h"
#include "m_shutdown.h"
#include "m_structs.h"
#include "m_thread.h"

#include "ms_parser.h"
#include "ms_socket.h"
#include "ms_globals.h"
#include "ms_mode.h"
#include "ms_msgout.h"
#include "ms_myclient.h"
#include "ms_network.h"
#include "ms_netutils.h"

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "misc.h"
#include "numnicks.h"
#include "typedefs.h"
#include "tokenizer.h"
#include "yxx.h"

struct irc_cmd_s {
#ifdef N2K_SUPPORT
    const char *tok;
#else
    const char *msg;
#endif
    irc_cmd_PP func;
};

typedef struct irc_cmd_s 	irc_cmd_s;

static irc_cmd_s irc_cmd_list[] = {
#ifdef N2K_SUPPORT
	{ TOK_PRIVATE, parse_privmsg },
	{ TOK_NICK, parse_nick },
	{ TOK_SERVER, parse_server },
	{ TOK_QUIT, parse_quit },
	{ TOK_SQUIT, parse_squit },
	{ TOK_KILL, parse_kill },
	{ TOK_PING, parse_ping },
	{ TOK_RPING, parse_rping },
	{ TOK_MODE, parse_mode },
	{ TOK_END_OF_BURST, parse_eb },
	{ TOK_END_OF_BURST_ACK, parse_ea }
	/* ,{ TOK_JOIN, parse_join } */
#else
	{ MSG_PRIVATE, parse_privmsg },
	{ MSG_NICK, parse_nick },
	{ MSG_SERVER, parse_server },
	{ MSG_QUIT, parse_quit },
	{ MSG_SQUIT, parse_squit },
	{ MSG_KILL, parse_kill },
	{ MSG_PING, parse_ping },
	{ MSG_RPING, parse_rping },
	{ MSG_MODE, parse_mode },
	{ MSG_END_OF_BURST, parse_eb },
	{ MSG_END_OF_BURST_ACK, parse_ea }
	/* ,{ MSG_JOIN, parse_join } */
#endif
};


extern long ipa_current;
extern long ipa_dups;
extern struct Server *mServer ;
extern long recvBytes;
extern htab *mHC;
extern htab *mHS;
extern server_handle_t mSA;

time_t start;
int msg_bogus = 0;
time_t burst_delay;
time_t startscan_time = 0;
extern time_t burst_starttime;;
extern pthread_cond_t sem_init;

server_handle_t mSArray;

htab *parstab = NULL;

void
parser_init()
{
   int i;
   
   parstab = jenkins_hcreate(2);
   
#define NUMCOMMANDS sizeof(irc_cmd_list)/sizeof(irc_cmd_s)
   
   /* Store commands (or tokens) in an hash table */
   
   for(i = 0; i < NUMCOMMANDS; i++)
#ifdef N2K_SUPPORT
      hadd(parstab, irc_cmd_list[i].tok, strlen(irc_cmd_list[i].tok),
         (void*)irc_cmd_list[i].func);
#else
      hadd(parstab, irc_cmd_list[i].msg, strlen(irc_cmd_list[i].msg),
         (void*)irc_cmd_list[i].func);
#endif
}

void
parser_destroy()
{
   if (parstab) {
      jenkins_hdestroy(parstab);
      parstab = NULL;
   }
}

void
irc_parse(char *msg)
{
   toktabptr ttab;

   stripline(msg);
   Debug((DBLBASIC, "[IN]: %s", msg));
   
   ttab = tokenize(msg, 11);
   if(ttab->size <= 1)
      return;
   
   if ((hcount(mHS) >= 1) && hfind(parstab, ttab->tok[1], strlen(ttab->tok[1]))) {
      
      /* Found. Call the appropriate command-handler function */
      
      (*(irc_cmd_PP)hstuff(parstab))(ttab); /* I can't do quicker, can i? :) */
      
   } else if (hcount(mHS) == 0) {
      yxx numyxx;
/*
  0                   1               2     3         4      5    6   7...
SERVER La-Defense-R.Fr.EU.IRCaide.org 1 900000000 985571276 J10 AFAH] :A Funky Speedy One!

non-n2k
SERVER ircu2.10.07.ircaide.org 1 960000000 1003670809 J10 GD] :A Funky Speedy One!
 */
      if( stricmp(ttab->tok[0], MSG_SERVER) == 0 ) {
         if(ttab->size < 8) {
            msg_bogus++;
            return;
         }
         time(&start);
      
         /* Deal with numeric/capacity as client_n/server_n */
         numyxx = yxx_to_int(ttab->tok[6]);
         MyUpNum = numyxx.server_n;
         MyUpCap = numyxx.client_n;
      
         server_add(MyNum, ttab->tok[1], MyUpNum, MyUpCap, 1 /* Bursting.. */) ;
	  
	     /* We are connected and are now a registered server, woohoo! */
         sendto_console("[%s] Running pxys %s", Me->id, VERSION) ;
         
      }
   }
   /* Ignore other messages... */
}


void
parse_privmsg(toktabptr ttab)
{
   if(ttab->size <= 3) {
      msg_bogus++;
      return;
   }
   
   if(!IsChannelName(ttab->tok[2]))
      myclient_handle_privmsg(ttab->tok[0], ttab->tok[3],
             untokenize(ttab, 4));
}


/*
   0  1  2  3     4      5            6                 7    8     9...
   Ak N NrG 4 985572592 NrG Quebec.QC.CA.IRCaide.Org B]AAAB AkAAJ :NrG
   
   0  1 2 3     4        5            6              7     8      9   10...
   Ao N E 2 000031337 cservice CMASTER.ircaide.org +oidk AAAAAA AoAAA :For help type..
   
     0   1   2      3
   AyAAE N test 985575361
   
   E NICK Nebula 2 1003542336 peedee crazy-horse.the-indian.net +m DPcmQU E0w :Nebula By MadMax
   
   GAE NICK mbu 1003683114
   
 */
void
parse_nick(toktabptr ttab)
{
   if(ttab->size == 4) {
                  /* Nick change */
      array_nickchange(ttab->tok[0], ttab->tok[2]);

   } else {
      int ident;
      
      if(ttab->size <= 9) {
         msg_bogus++;
         return;
      }
      
      if (*ttab->tok[5] == '~')
         ident = 0;
      else
         ident = 1;
      
      if(*ttab->tok[7] == '+') {
         array_addclient(ttab->tok[2], ttab->tok[7], ttab->tok[8],
      	      ttab->tok[9], ident);
      }
      else {
         array_addclient(ttab->tok[2], NULL, ttab->tok[7],
      	      ttab->tok[8], ident);
      }
   }
}

/*
0  1         2             3 4     5      6    7   8 9...
AF S VDA.FR.EU.IRCAide.Org 2 0 984087546 J10 ARAD] 0 :[Jeb ircaide.org server]
AF S services.EU.IRCaide.Org 2 0 985635871 P10 XD] 0 :[ecp-net.com IRC Service]  

--disable-n2k
G SERVER ircu2.10.07.bis.ircaide.org 2 0 1003671699 P10 ID] 0 :A Funky Speedy One!
   
*/

void
parse_server(toktabptr ttab)
{
   if(ttab->size <= 9) {
      msg_bogus++;
   } else {
      unsigned int numfrom = base64toint(ttab->tok[0]);
      char *t1 = ttab->tok[6] + 1;
      
      if(atoi(t1) >= 10) {
         yxx numyxx = yxx_to_int(ttab->tok[7]);
         server_add( numfrom, ttab->tok[2],
                     numyxx.server_n, numyxx.client_n, (*ttab->tok[6] == 'J'));
      } else {
         log_error("parse_server> %s: P9 unsupported", ttab->tok[2]);
         fprintf(stderr, "parse_server> %s: P9 unsupported\n", ttab->tok[2]);
         m_shutdown();
      }
   }
}

/*
 * :Anubis QUIT :SlipKnot OwNz YoU aLl! Korn, SlipKnot RoXoR!
 */
void
parse_quit(toktabptr ttab)
{
#ifdef N2K_SUPPORT

   array_delclient(ttab->tok[0]);
   
#else
/* That's bad, but non-n2k sucks */

   char charYXX[4];
   struct Client* cptr;
   
   cptr = array_findclient_from_nick(ttab->tok[0] + 1);
   
   if (cptr) {
      inttobase64(charYXX, cptr->nnick, 1);
      inttobase64(charYXX + 1, cptr->nserv, 2);
      array_delclient(charYXX);
   }
#endif
}

/*
  0      1              2                3     4...
:mbuna SQUIT Paris.Fr.EU.IRCaide.org 983950525 :t e s t
CJAAH SQ hub7.eu.bugged.org 0 :blah
*/
void
parse_squit(toktabptr ttab)
{
   if(ttab->size <= 4) {
      msg_bogus++;
      return;
   }
   do_server_quit(ttab->tok[2]);
}

/*
  0   1   2                                      3                            4...
AEACR D A3ADS :La-Defense.Fr.EU.IRCaide.org!ppp10-240.lino.sympatico.ca!MyLlY (tas tu honte?)

!n2k
:mbuna KILL IAA :ca-ol-bordeaux-10-201.abo.wanadoo.fr!mbuna (test)
 */
 
void
parse_kill(toktabptr ttab)
{
   if(ttab->size <= 3) {
      msg_bogus++;
      return;
   }
   array_delclient(ttab->tok[2]);
}


void
parse_ping(toktabptr ttab)
{
   unsigned int yy_int;
                              /* Only reply to pings from our uplink. */
   yy_int = base64toint(ttab->tok[0]);
   if (yy_int != MyUpNum)
      return;
   
   if(ttab->size >= 3)
#ifdef N2K_SUPPORT
      sendto_uplink("%s %s %s", MyYY, TOK_PONG, ttab->tok[2]);
#else
      sendto_uplink("%s %s %s", MyYY, MSG_PONG, ttab->tok[2]);
#endif
   else
#ifdef N2K_SUPPORT
      sendto_uplink("%s %s :%s", MyYY, TOK_PONG, Me->name);
#else
      sendto_uplink("%s %s :%s", MyYY, MSG_PONG, Me->name);
#endif
}

/*
 * [IN]: :ircu2.10.07.ircaide.org RPING R mbuna 1003681380 30195 :<No client start time>
 */
void
parse_rping(toktabptr ttab)
{
   if (ttab->size <= 6)
      msg_bogus++;
   
   else {
      char timebuf[18];
#ifdef N2K_SUPPORT
      struct Server *sptr;
#endif
      
      militime_r(ttab->tok[4], ttab->tok[5], timebuf, sizeof(timebuf));
      
#ifdef N2K_SUPPORT
      sptr = (struct Server *) mSArray[base64toint(ttab->tok[0])];
      if (sptr != NULL)
         sendto_uplink("%s %s %s %s %s %s %s",
            MyYY, TOK_RPONG, sptr->name, ttab->tok[3], ttab->tok[4],
            timebuf, ttab->tok[6]);
#else
         sendto_uplink(":%s %s %s %s %s %s %s",
            Me->name, MSG_RPONG, ttab->tok[0] + 1, ttab->tok[3], ttab->tok[4],
            timebuf, ttab->tok[6]);
#endif /* N2K_SUPPORT */
   }
}


/*
  0   1   2      3    (4...)
AEACR M #funky +m-iko test AyAAF  

  0   1   2      3
BLAAG M KickuP :+iw

 */
 
void
parse_mode(toktabptr ttab)
{
   if(ttab->size <= 3)
      msg_bogus++;
   else {
                         /* Ignore channel modes changes. */
      if(!IsChannelName(ttab->tok[2]))
         ms_mode(ttab->tok[0], (*ttab->tok[3] == ':') ? (ttab->tok[3]+1) : ttab->tok[3]);
   }
}


void
parse_eb(toktabptr ttab)
{
   unsigned int yy_int;
                              /* Only reply to EB from our uplink. */
   yy_int = base64toint(ttab->tok[0]);
   if (yy_int != MyUpNum)
      return;
   
   time(&burst_delay); /* Reentrant */
   burst_delay -= burst_starttime;
   
#ifdef N2K_SUPPORT
   sendto_uplink("%s %s", MyYY, TOK_END_OF_BURST_ACK);
#else
   sendto_uplink("%s %s", MyYY, MSG_END_OF_BURST_ACK);
#endif
}


void
parse_ea(toktabptr ttab)
{
   unsigned int yy_int;
   struct Server *sptr;
                              
   yy_int = base64toint(ttab->tok[0]);
   
   if ((sptr = mSA[yy_int]))
      sptr->bursting = 0;
   
   /* Only reply to EA from our uplink. */
   if (yy_int != MyUpNum)
      return;
   
   /* Ack from our uplink */
   sendto_console("[%s] Loaded %ld clients in %ld secs (%ldK - %4.2fK/s)",
              Me->id, ipa_current, burst_delay,
              recvBytes/1024,
              burst_delay ? (recvBytes/1024.)/burst_delay : recvBytes/1024.) ;
   
   time(&startscan_time);
   
   Debug((DBLMAX, "parse_ea ok"));
   
   /* Let's go. */
   scanner_startall();
   
   Debug((DBLMAX, "scanner_startall ok"));
}

/*
  0   1     2
AEACS J #proxy_info
*/

/* Off to save CPU
void
parse_join(toktabptr ttab)
{
#if 0 
   if(stricmp(ttab->tok[2], MyConsole->channel) == 0)
      myclient_handle_join(MyConsole->channel, ttab->tok[0]);
#endif
}
 */

/* ms_parse.c */
