/* Pxys - An IRC ProxyScan, strnlwrcpy.c
 * 
 * File: strncpy.c
 * Author: James Hall (jch1003@cl.cam.ac.uk)
 * Copyright (C) University of Cambridge Computer Laboratory, 1994
 * PACKAGE: Nemesis C library
 * 
 * strnlwrcpy.c
 * Copyright (C) 2001 Stephane Thiell <mbuna@bugged.org>
 *
 * FUNCTION: Copy nn (<= n) bytes from s2 to s1 with lowercase conversion - return s1
 *           - if strlen(s2) < n:  append null chars to make strlen(s1) == n
 *           - if strlen(s2) >= n: copy exactly n chars (ie. no terminator!)

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * strnlwrcpy() is a "merge" of strncpy() and strlwr() (strtolower)
 *
 * $Id: strnlwrcpy.c,v 1.1 2001/10/31 20:49:19 mbuna Exp $
 */

#include "config.h"
#include <string.h>
#include <ctype.h>

#include "strnlwrcpy.h"

char*
strnlwrcpy(char *s1, const char *s2, size_t n)
{
   register char c;
   char *s = s1;
   
   --s1;
   
   if (n >= 4) {
      size_t n4 = n >> 2;
   
      for (;;) {
         c = *s2++;
         *++s1 = tolower(c);
         if (c == '\0')
            break;
         c = *s2++;
         *++s1 = tolower(c);
         if (c == '\0')
            break;
         c = *s2++;
         *++s1 = tolower(c);
         if (c == '\0')
            break;
          c = *s2++;
         *++s1 = tolower(c);
         if (c == '\0')
            break;
         if (--n4 == 0)
            goto last_chars;
      }
      n = n - (s1 - s) - 1;
      if (n == 0)
         return s;
      goto zero_fill;
   }
   
last_chars:
   n &= 3;
   if (n == 0)
      return s;
   
   do {
      c = *s2++;
      *++s1 = tolower(c);
      if (--n == 0)
         return s;
   } while (c != '\0');
   
zero_fill:
   do
      *++s1 = '\0';
   while (--n > 0);
   
   return s;
}


/* strnlwrcpy.c */
